@@ -1,6 +1,6 @@
 
 /*
- * $Id: forward.c,v 1.82.2.3 2003/03/17 18:34:19 hno Exp $
+ * $Id: forward.c,v 1.115 2006/06/30 21:23:05 hno Exp $
  *
  * DEBUG: section 17    Request Forwarding
  * AUTHOR: Duane Wessels
@@ -36,12 +36,20 @@
 
 #include "squid.h"
 
+#if LINUX_NETFILTER
+#include <linux/netfilter_ipv4.h>
+#endif
+#if LINUX_TPROXY
+#include <linux/netfilter_ipv4/ip_tproxy.h>
+#endif
+
 static PSC fwdStartComplete;
 static void fwdDispatch(FwdState *);
 static void fwdConnectStart(void *);	/* should be same as EVH */
 static void fwdStateFree(FwdState * fwdState);
 static PF fwdConnectTimeout;
 static PF fwdServerClosed;
+static PF fwdPeerClosed;
 static CNCB fwdConnectDone;
 static int fwdCheckRetry(FwdState * fwdState);
 static int fwdReforward(FwdState *);
@@ -102,6 +110,8 @@ fwdStateFree(FwdState * fwdState)
 	    storeReleaseRequest(e);
 	}
     }
+    if (EBIT_TEST(e->flags, ENTRY_DEFER_READ))
+	storeResetDefer(e);
     if (storePendingNClients(e) > 0)
 	assert(!EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT));
     p = fwdStateServerPeer(fwdState);
@@ -110,7 +120,7 @@ fwdStateFree(FwdState * fwdState)
     fwdState->request = NULL;
     if (fwdState->err)
 	errorStateFree(fwdState->err);
-    storeUnregisterAbort(e);
+    storeClientUnregisterAbort(e);
     storeUnlockObject(e);
     fwdState->entry = NULL;
     sfd = fwdState->server_fd;
@@ -119,8 +129,6 @@ fwdStateFree(FwdState * fwdState)
 	fwdState->server_fd = -1;
 	debug(17, 3) ("fwdStateFree: closing FD %d\n", sfd);
 	comm_close(sfd);
-	if (p)
-	    p->stats.conn_open--;
     }
     cbdataFree(fwdState);
 }
@@ -136,7 +144,9 @@ fwdCheckRetry(FwdState * fwdState)
 	return 0;
     if (fwdState->n_tries > 10)
 	return 0;
-    if (squid_curtime - fwdState->start > Config.Timeout.connect)
+    if (fwdState->origin_tries > 2)
+	return 0;
+    if (squid_curtime - fwdState->start >= Config.Timeout.forward)
 	return 0;
     if (fwdState->flags.dont_retry)
 	return 0;
@@ -151,7 +161,7 @@ fwdCheckRetriable(FwdState * fwdState)
     /* If there is a request body then Squid can only try once
      * even if the method is indempotent
      */
-    if (fwdState->request->body_connection)
+    if (fwdState->request->body_reader)
 	return 0;
 
     /* RFC2616 9.1 Safe and Idempotent Methods */
@@ -179,7 +189,10 @@ fwdServerClosed(int fd, void *data)
     debug(17, 2) ("fwdServerClosed: FD %d %s\n", fd, storeUrl(fwdState->entry));
     assert(fwdState->server_fd == fd);
     fwdState->server_fd = -1;
+    if (EBIT_TEST(fwdState->entry->flags, ENTRY_DEFER_READ))
+	storeResetDefer(fwdState->entry);
     if (fwdCheckRetry(fwdState)) {
+	int originserver = (fwdState->servers->peer == NULL);
 	debug(17, 3) ("fwdServerClosed: re-forwarding (%d tries, %d secs)\n",
 	    fwdState->n_tries,
 	    (int) (squid_curtime - fwdState->start));
@@ -197,10 +210,11 @@ fwdServerClosed(int fd, void *data)
 		/* Use next. The last "direct" entry is retried multiple times */
 		fwdState->servers = fs->next;
 		fwdServerFree(fs);
+		originserver = 0;
 	    }
 	}
-	/* use eventAdd to break potential call sequence loops */
-	eventAdd("fwdConnectStart", fwdConnectStart, fwdState, 0.0, 0);
+	/* use eventAdd to break potential call sequence loops and to slow things down a little */
+	eventAdd("fwdConnectStart", fwdConnectStart, fwdState, originserver ? 0.05 : 0.005, 0);
 	return;
     }
     if (!fwdState->err && shutting_down) {
@@ -210,17 +224,119 @@ fwdServerClosed(int fd, void *data)
     fwdStateFree(fwdState);
 }
 
+#if USE_SSL
+static void
+fwdNegotiateSSL(int fd, void *data)
+{
+    FwdState *fwdState = data;
+    FwdServer *fs = fwdState->servers;
+    SSL *ssl = fd_table[fd].ssl;
+    int ret;
+    ErrorState *err;
+    request_t *request = fwdState->request;
+
+    errno = 0;
+    ERR_clear_error();
+    if ((ret = SSL_connect(ssl)) <= 0) {
+	int ssl_error = SSL_get_error(ssl, ret);
+	switch (ssl_error) {
+	case SSL_ERROR_WANT_READ:
+	    commSetSelect(fd, COMM_SELECT_READ, fwdNegotiateSSL, fwdState, 0);
+	    return;
+	case SSL_ERROR_WANT_WRITE:
+	    commSetSelect(fd, COMM_SELECT_WRITE, fwdNegotiateSSL, fwdState, 0);
+	    return;
+	default:
+	    debug(81, 1) ("fwdNegotiateSSL: Error negotiating SSL connection on FD %d: %s (%d/%d/%d)\n", fd, ERR_error_string(ERR_get_error(), NULL), ssl_error, ret, errno);
+	    err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
+#ifdef EPROTO
+	    err->xerrno = EPROTO;
+#else
+	    err->xerrno = EACCES;
+#endif
+	    err->request = requestLink(request);
+	    fwdFail(fwdState, err);
+	    if (fs->peer) {
+		peerConnectFailed(fs->peer);
+		fs->peer->stats.conn_open--;
+	    }
+	    comm_close(fd);
+	    return;
+	}
+    }
+    if (fs->peer && !SSL_session_reused(ssl)) {
+	if (fs->peer->sslSession)
+	    SSL_SESSION_free(fs->peer->sslSession);
+	fs->peer->sslSession = SSL_get1_session(ssl);
+    }
+#if NOT_YET
+    if (verify_domain) {
+	char *host;
+	STACK_OF(GENERAL_NAME) * altnames;
+	if (fs->peer) {
+	    if (fs->peer->ssldomain)
+		host = fs->peer->ssldomain;
+	    else
+		host = fs->peer->host;
+	} else {
+	    host = fs->request->host;
+	}
+	if (!ssl_verify_domain(host, ssl)) {
+	    debug(17, 1) ("Warning: SSL certificate does not match host name '%s'\n", host);
+	}
+    }
+#endif
+    fwdDispatch(fwdState);
+}
+
+static void
+fwdInitiateSSL(FwdState * fwdState)
+{
+    FwdServer *fs = fwdState->servers;
+    int fd = fwdState->server_fd;
+    SSL *ssl;
+    SSL_CTX *sslContext = NULL;
+    peer *peer = fs->peer;
+    if (peer) {
+	assert(peer->use_ssl);
+	sslContext = peer->sslContext;
+    } else {
+	sslContext = Config.ssl_client.sslContext;
+    }
+    assert(sslContext);
+    if ((ssl = SSL_new(sslContext)) == NULL) {
+	ErrorState *err;
+	debug(83, 1) ("fwdInitiateSSL: Error allocating handle: %s\n",
+	    ERR_error_string(ERR_get_error(), NULL));
+	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR);
+	err->xerrno = errno;
+	err->request = requestLink(fwdState->request);
+	fwdFail(fwdState, err);
+	fwdStateFree(fwdState);
+	return;
+    }
+    SSL_set_fd(ssl, fd);
+    if (peer) {
+	if (peer->sslSession)
+	    SSL_set_session(ssl, peer->sslSession);
+    }
+    fd_table[fd].ssl = ssl;
+    fd_table[fd].read_method = &ssl_read_method;
+    fd_table[fd].write_method = &ssl_write_method;
+    fwdNegotiateSSL(fd, fwdState);
+}
+#endif
+
 static void
 fwdConnectDone(int server_fd, int status, void *data)
 {
     FwdState *fwdState = data;
-    static FwdState *current = NULL;
     FwdServer *fs = fwdState->servers;
     ErrorState *err;
     request_t *request = fwdState->request;
-    assert(current != fwdState);
-    current = fwdState;
     assert(fwdState->server_fd == server_fd);
+    if (Config.onoff.log_ip_on_direct && status != COMM_ERR_DNS && fs->code == HIER_DIRECT)
+	hierarchyNote(&fwdState->request->hier, fs->code, fd_table[server_fd].ipaddr);
     if (status == COMM_ERR_DNS) {
 	/*
 	 * Only set the dont_retry flag if the DNS lookup fails on
@@ -233,44 +349,31 @@ fwdConnectDone(int server_fd, int status, void *data)
 	    request->host);
 	err = errorCon(ERR_DNS_FAIL, HTTP_SERVICE_UNAVAILABLE);
 	err->dnsserver_msg = xstrdup(dns_error_message);
-	err->request = requestLink(request);
 	fwdFail(fwdState, err);
-	if (fs->peer)
-	    fs->peer->stats.conn_open--;
 	comm_close(server_fd);
     } else if (status != COMM_OK) {
 	assert(fs);
 	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
 	err->xerrno = errno;
-	if (fs->peer) {
-	    err->host = xstrdup(fs->peer->host);
-	    err->port = fs->peer->http_port;
-	} else {
-	    err->host = xstrdup(request->host);
-	    err->port = request->port;
-	}
-	err->request = requestLink(request);
 	fwdFail(fwdState, err);
-	if (fs->peer) {
+	if (fs->peer)
 	    peerConnectFailed(fs->peer);
-	    fs->peer->stats.conn_open--;
-	}
 	comm_close(server_fd);
     } else {
 	debug(17, 3) ("fwdConnectDone: FD %d: '%s'\n", server_fd, storeUrl(fwdState->entry));
-	if (fs->peer)
-	    hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
-	else if (Config.onoff.log_ip_on_direct)
-	    hierarchyNote(&fwdState->request->hier, fs->code, fd_table[server_fd].ipaddr);
-	else
-	    hierarchyNote(&fwdState->request->hier, fs->code, request->host);
 	fd_note(server_fd, storeUrl(fwdState->entry));
 	fd_table[server_fd].uses++;
-	if (fs->peer)
+	if (fd_table[server_fd].uses == 1 && fs->peer)
 	    peerConnectSucceded(fs->peer);
+#if USE_SSL
+	if ((fs->peer && fs->peer->use_ssl) ||
+	    (!fs->peer && request->protocol == PROTO_HTTPS)) {
+	    fwdInitiateSSL(fwdState);
+	    return;
+	}
+#endif
 	fwdDispatch(fwdState);
     }
-    current = NULL;
 }
 
 static void
@@ -278,13 +381,14 @@ fwdConnectTimeout(int fd, void *data)
 {
     FwdState *fwdState = data;
     StoreEntry *entry = fwdState->entry;
+    FwdServer *fs = fwdState->servers;
     ErrorState *err;
-    peer *p = fwdStateServerPeer(fwdState);
     debug(17, 2) ("fwdConnectTimeout: FD %d: '%s'\n", fd, storeUrl(entry));
     assert(fd == fwdState->server_fd);
+    if (Config.onoff.log_ip_on_direct && fs->code == HIER_DIRECT && fd_table[fd].ipaddr[0])
+	hierarchyNote(&fwdState->request->hier, fs->code, fd_table[fd].ipaddr);
     if (entry->mem_obj->inmem_hi == 0) {
 	err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT);
-	err->request = requestLink(fwdState->request);
 	err->xerrno = ETIMEDOUT;
 	fwdFail(fwdState, err);
 	/*
@@ -294,8 +398,6 @@ fwdConnectTimeout(int fd, void *data)
 	    if (fwdState->servers->peer)
 		peerConnectFailed(fwdState->servers->peer);
     }
-    if (p)
-	p->stats.conn_open--;
     comm_close(fd);
 }
 
@@ -360,37 +462,92 @@ fwdConnectStart(void *data)
     ErrorState *err;
     FwdServer *fs = fwdState->servers;
     const char *host;
+    const char *name;
     unsigned short port;
-    time_t ctimeout;
+    const char *domain = NULL;
+    int ctimeout;
+    int ftimeout = Config.Timeout.forward - (squid_curtime - fwdState->start);
     struct in_addr outgoing;
     unsigned short tos;
+#if LINUX_TPROXY
+    struct in_tproxy itp;
+#endif
+
     assert(fs);
     assert(fwdState->server_fd == -1);
     debug(17, 3) ("fwdConnectStart: %s\n", url);
     if (fs->peer) {
 	host = fs->peer->host;
+	name = fs->peer->name;
 	port = fs->peer->http_port;
+	if (fs->peer->options.originserver)
+	    domain = fwdState->request->host;
 	ctimeout = fs->peer->connect_timeout > 0 ? fs->peer->connect_timeout
 	    : Config.Timeout.peer_connect;
-    } else if (fwdState->request->flags.accelerated &&
-	Config.Accel.single_host && Config.Accel.host) {
-	host = Config.Accel.host;
-	port = Config.Accel.port;
-	ctimeout = Config.Timeout.connect;
     } else {
-	host = fwdState->request->host;
+	host = name = fwdState->request->host;
 	port = fwdState->request->port;
 	ctimeout = Config.Timeout.connect;
     }
-    if (fwdCheckRetriable(fwdState)) {
-	if ((fd = pconnPop(host, port)) >= 0) {
-	    debug(17, 3) ("fwdConnectStart: reusing pconn FD %d\n", fd);
+    if (ftimeout < 0)
+	ftimeout = 5;
+    if (ftimeout < ctimeout)
+	ctimeout = ftimeout;
+    fwdState->request->flags.pinned = 0;
+    if (fs->code == PINNED) {
+	int auth;
+	fd = clientGetPinnedConnection(fwdState->request->pinned_connection, fwdState->request, fs->peer, &auth);
+	if (fd >= 0) {
+#if 0
+	    if (!fs->peer)
+		fs->code = HIER_DIRECT;
+#endif
 	    fwdState->server_fd = fd;
 	    fwdState->n_tries++;
+	    fwdState->request->flags.pinned = 1;
+	    if (auth)
+		fwdState->request->flags.auth = 1;
 	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
 	    fwdConnectDone(fd, COMM_OK, fwdState);
 	    return;
 	}
+	/* Failure. Fall back on next path */
+	cbdataUnlock(fwdState->request->pinned_connection);
+	fwdState->request->pinned_connection = NULL;
+	fwdState->servers = fs->next;
+	fwdServerFree(fs);
+	fwdConnectStart(fwdState);
+	return;
+    }
+#if LINUX_TPROXY
+    if (fd == -1 && fwdState->request->flags.tproxy)
+	fd = pconnPop(name, port, domain, &fwdState->request->client_addr, 0);
+#endif
+    if (fd == -1)
+	fd = pconnPop(name, port, domain, NULL, 0);
+    if (fd != -1) {
+	if (fwdCheckRetriable(fwdState)) {
+	    debug(17, 3) ("fwdConnectStart: reusing pconn FD %d\n", fd);
+	    fwdState->server_fd = fd;
+	    fwdState->n_tries++;
+	    if (!fs->peer)
+		fwdState->origin_tries++;
+	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
+	    if (fs->peer)
+		hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
+	    else if (Config.onoff.log_ip_on_direct && fs->code == HIER_DIRECT)
+		hierarchyNote(&fwdState->request->hier, fs->code, fd_table[fd].ipaddr);
+	    else
+		hierarchyNote(&fwdState->request->hier, fs->code, name);
+	    fwdConnectDone(fd, COMM_OK, fwdState);
+	    return;
+	} else {
+	    /* Discard the persistent connection to not cause
+	     * a imbalance in number of conenctions open if there
+	     * is a lot of POST requests
+	     */
+	    comm_close(fd);
+	}
     }
 #if URL_CHECKSUM_DEBUG
     assert(fwdState->entry->mem_obj->chksum == url_checksum(url));
@@ -401,7 +558,7 @@ fwdConnectStart(void *data)
     debug(17, 3) ("fwdConnectStart: got addr %s, tos %d\n",
 	inet_ntoa(outgoing), tos);
     fd = comm_openex(SOCK_STREAM,
-	0,
+	IPPROTO_TCP,
 	outgoing,
 	0,
 	COMM_NONBLOCKING,
@@ -411,26 +568,60 @@ fwdConnectStart(void *data)
 	debug(50, 4) ("fwdConnectStart: %s\n", xstrerror());
 	err = errorCon(ERR_SOCKET_FAILURE, HTTP_INTERNAL_SERVER_ERROR);
 	err->xerrno = errno;
-	err->request = requestLink(fwdState->request);
 	fwdFail(fwdState, err);
 	fwdStateFree(fwdState);
 	return;
     }
     fwdState->server_fd = fd;
     fwdState->n_tries++;
+    if (!fs->peer)
+	fwdState->origin_tries++;
     /*
      * stats.conn_open is used to account for the number of
      * connections that we have open to the peer, so we can limit
      * based on the max-conn option.  We need to increment here,
      * even if the connection may fail.
      */
-    if (fs->peer)
+    if (fs->peer) {
 	fs->peer->stats.conn_open++;
+	comm_add_close_handler(fd, fwdPeerClosed, fs->peer);
+    }
     comm_add_close_handler(fd, fwdServerClosed, fwdState);
     commSetTimeout(fd,
 	ctimeout,
 	fwdConnectTimeout,
 	fwdState);
+    if (fs->peer) {
+	hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
+    } else {
+#if LINUX_TPROXY
+	if (fwdState->request->flags.tproxy) {
+
+	    itp.v.addr.faddr.s_addr = fwdState->src.sin_addr.s_addr;
+	    itp.v.addr.fport = 0;
+
+	    /* If these syscalls fail then we just fallback to connecting
+	     * normally by simply ignoring the errors...
+	     */
+	    itp.op = TPROXY_ASSIGN;
+	    if (setsockopt(fd, SOL_IP, IP_TPROXY, &itp, sizeof(itp)) == -1) {
+		debug(20, 1) ("tproxy ip=%s,0x%x,port=%d ERROR ASSIGN\n",
+		    inet_ntoa(itp.v.addr.faddr),
+		    itp.v.addr.faddr.s_addr,
+		    itp.v.addr.fport);
+	    } else {
+		itp.op = TPROXY_FLAGS;
+		itp.v.flags = ITP_CONNECT;
+		if (setsockopt(fd, SOL_IP, IP_TPROXY, &itp, sizeof(itp)) == -1) {
+		    debug(20, 1) ("tproxy ip=%x,port=%d ERROR CONNECT\n",
+			itp.v.addr.faddr.s_addr,
+			itp.v.addr.fport);
+		}
+	    }
+	}
+#endif
+	hierarchyNote(&fwdState->request->hier, fs->code, fwdState->request->host);
+    }
     commConnectStart(fd, host, port, fwdConnectDone, fwdState);
 }
 
@@ -453,7 +644,6 @@ fwdStartFail(FwdState * fwdState)
     ErrorState *err;
     debug(17, 3) ("fwdStartFail: %s\n", storeUrl(fwdState->entry));
     err = errorCon(ERR_CANNOT_FORWARD, HTTP_SERVICE_UNAVAILABLE);
-    err->request = requestLink(fwdState->request);
     err->xerrno = errno;
     fwdFail(fwdState, err);
     fwdStateFree(fwdState);
@@ -466,28 +656,37 @@ fwdDispatch(FwdState * fwdState)
     request_t *request = fwdState->request;
     StoreEntry *entry = fwdState->entry;
     ErrorState *err;
+    int server_fd = fwdState->server_fd;
     debug(17, 3) ("fwdDispatch: FD %d: Fetching '%s %s'\n",
 	fwdState->client_fd,
 	RequestMethodStr[request->method],
 	storeUrl(entry));
-    /*assert(!EBIT_TEST(entry->flags, ENTRY_DISPATCHED)); */
-    assert(entry->ping_status != PING_WAITING);
-    assert(entry->lock_count);
-    EBIT_SET(entry->flags, ENTRY_DISPATCHED);
-    netdbPingSite(request->host);
     /*
      * Assert that server_fd is set.  This is to guarantee that fwdState
      * is attached to something and will be deallocated when server_fd
      * is closed.
      */
-    assert(fwdState->server_fd > -1);
+    assert(server_fd > -1);
+    /*assert(!EBIT_TEST(entry->flags, ENTRY_DISPATCHED)); */
+    assert(entry->ping_status != PING_WAITING);
+    assert(entry->lock_count);
+    EBIT_SET(entry->flags, ENTRY_DISPATCHED);
+    netdbPingSite(request->host);
+    entry->mem_obj->refresh_timestamp = squid_curtime;
     if (fwdState->servers && (p = fwdState->servers->peer)) {
 	p->stats.fetches++;
 	fwdState->request->peer_login = p->login;
+	fwdState->request->peer_domain = p->domain;
 	httpStart(fwdState);
     } else {
 	fwdState->request->peer_login = NULL;
+	fwdState->request->peer_domain = NULL;
 	switch (request->protocol) {
+#if USE_SSL
+	case PROTO_HTTPS:
+	    httpStart(fwdState);
+	    break;
+#endif
 	case PROTO_HTTP:
 	    httpStart(fwdState);
 	    break;
@@ -512,7 +711,6 @@ fwdDispatch(FwdState * fwdState)
 	    debug(17, 1) ("fwdDispatch: Cannot retrieve '%s'\n",
 		storeUrl(entry));
 	    err = errorCon(ERR_UNSUP_REQ, HTTP_BAD_REQUEST);
-	    err->request = requestLink(request);
 	    fwdFail(fwdState, err);
 	    /*
 	     * Force a persistent connection to be closed because
@@ -525,11 +723,6 @@ fwdDispatch(FwdState * fwdState)
 	     * transient (network) error; its a bug.
 	     */
 	    fwdState->flags.dont_retry = 1;
-	    /*
-	     * this assertion exists because if we are connected to
-	     * a peer, then we need to decrement p->stats.conn_open.
-	     */
-	    assert(NULL == p);
 	    comm_close(fwdState->server_fd);
 	    break;
 	}
@@ -554,6 +747,8 @@ fwdReforward(FwdState * fwdState)
     }
     if (fwdState->n_tries > 9)
 	return 0;
+    if (fwdState->origin_tries > 1)
+	return 0;
     if (fwdState->request->flags.body_sent)
 	return 0;
     assert(fs);
@@ -580,6 +775,29 @@ fwdServersFree(FwdServer ** FS)
     }
 }
 
+void
+fwdStartPeer(peer * p, StoreEntry * e, request_t * r)
+{
+    FwdState *fwdState;
+    FwdServer *peer = NULL;
+    debug(17, 3) ("fwdStartPeer: '%s'\n", storeUrl(e));
+    e->mem_obj->request = requestLink(r);
+#if URL_CHECKSUM_DEBUG
+    assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
+#endif
+    fwdState = cbdataAlloc(FwdState);
+    fwdState->entry = e;
+    fwdState->client_fd = -1;
+    fwdState->server_fd = -1;
+    fwdState->request = requestLink(r);
+    fwdState->start = squid_curtime;
+    storeLockObject(e);
+    EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
+    storeRegisterAbort(e, fwdAbort, fwdState);
+    peerAddFwdServer(&peer, p, HIER_DIRECT);
+    fwdStartComplete(peer, fwdState);
+}
+
 void
 fwdStart(int fd, StoreEntry * e, request_t * r)
 {
@@ -592,7 +810,7 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
      * from peer_digest.c, asn.c, netdb.c, etc and should always
      * be allowed.  yuck, I know.
      */
-    if (r->client_addr.s_addr != no_addr.s_addr) {
+    if (r->client_addr.s_addr != no_addr.s_addr && r->protocol != PROTO_INTERNAL && r->protocol != PROTO_CACHEOBJ) {
 	/*      
 	 * Check if this host is allowed to fetch MISSES from us (miss_access)
 	 */
@@ -603,7 +821,11 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
 	ch.request = r;
 	answer = aclCheckFast(Config.accessList.miss, &ch);
 	if (answer == 0) {
-	    err = errorCon(ERR_FORWARDING_DENIED, HTTP_FORBIDDEN);
+	    err_type page_id;
+	    page_id = aclGetDenyInfoPage(&Config.denyInfoList, AclMatchedName);
+	    if (page_id == ERR_NONE)
+		page_id = ERR_FORWARDING_DENIED;
+	    err = errorCon(page_id, HTTP_FORBIDDEN);
 	    err->request = requestLink(r);
 	    err->src_addr = r->client_addr;
 	    errorAppendEntry(e, err);
@@ -612,7 +834,6 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
     }
     debug(17, 3) ("fwdStart: '%s'\n", storeUrl(e));
     e->mem_obj->request = requestLink(r);
-    e->mem_obj->fd = fd;
 #if URL_CHECKSUM_DEBUG
     assert(e->mem_obj->chksum == url_checksum(e->mem_obj->url));
 #endif
@@ -645,6 +866,15 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
     fwdState->server_fd = -1;
     fwdState->request = requestLink(r);
     fwdState->start = squid_curtime;
+
+#if LINUX_TPROXY
+    /* If we need to transparently proxy the request
+     * then we need the client source address and port */
+    fwdState->src.sin_family = AF_INET;
+    fwdState->src.sin_addr = r->client_addr;
+    fwdState->src.sin_port = r->client_port;
+#endif
+
     storeLockObject(e);
     EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
     storeRegisterAbort(e, fwdAbort, fwdState);
@@ -669,8 +899,14 @@ fwdCheckDeferRead(int fd, void *data)
 	(void) 0;
     else {
 	int i = delayMostBytesWanted(mem, INT_MAX);
-	if (0 == i)
+	if (0 == i) {
+	    /* No storeDeferRead here as it's not store dependent.
+	     * Will get changed the day delay pools interact nicely
+	     * with the store..
+	     */
+	    commDeferFD(fd);
 	    return 1;
+	}
 	/* was: rc = -(rc != INT_MAX); */
 	else if (INT_MAX == i)
 	    rc = 0;
@@ -678,17 +914,33 @@ fwdCheckDeferRead(int fd, void *data)
 	    rc = -1;
     }
 #endif
+    if (EBIT_TEST(e->flags, ENTRY_DEFER_READ)) {
+	commDeferFD(mem->serverfd);
+	return 1;
+    }
     if (EBIT_TEST(e->flags, ENTRY_FWD_HDR_WAIT))
 	return rc;
-    if (mem->inmem_hi - storeLowestMemReaderOffset(e) < READ_AHEAD_GAP)
-	return rc;
-    return 1;
+    if (EBIT_TEST(e->flags, RELEASE_REQUEST)) {
+	/* Just a small safety cap to defer storing more data into the object
+	 * if there already is way too much. This handles the case when there
+	 * is disk clients pending on a too large object being fetched and a
+	 * few other corner cases.
+	 */
+	if (fd >= 0 && mem->inmem_hi - mem->inmem_lo > SM_PAGE_SIZE + Config.Store.maxInMemObjSize + Config.readAheadGap) {
+	    storeDeferRead(e, fd);
+	    return 1;
+	}
+    }
+    if (fd >= 0 && mem->inmem_hi - storeLowestMemReaderOffset(e) > Config.readAheadGap) {
+	storeDeferRead(e, fd);
+	return 1;
+    }
+    return rc;
 }
 
 void
 fwdFail(FwdState * fwdState, ErrorState * errorState)
 {
-    assert(EBIT_TEST(fwdState->entry->flags, ENTRY_FWD_HDR_WAIT));
     debug(17, 3) ("fwdFail: %s \"%s\"\n\t%s\n",
 	err_type_str[errorState->type],
 	httpStatusString(errorState->http_status),
@@ -696,6 +948,8 @@ fwdFail(FwdState * fwdState, ErrorState * errorState)
     if (fwdState->err)
 	errorStateFree(fwdState->err);
     fwdState->err = errorState;
+    if (!errorState->request)
+	errorState->request = requestLink(fwdState->request);
 }
 
 /*
@@ -709,6 +963,16 @@ fwdAbort(void *data)
     fwdStateFree(fwdState);
 }
 
+/*
+ * Accounts for closed persistent connections
+ */
+static void
+fwdPeerClosed(int fd, void *data)
+{
+    peer *p = data;
+    p->stats.conn_open--;
+}
+
 /*
  * Frees fwdState without closing FD or generating an abort
  */
@@ -718,6 +982,9 @@ fwdUnregister(int fd, FwdState * fwdState)
     debug(17, 3) ("fwdUnregister: %s\n", storeUrl(fwdState->entry));
     assert(fd == fwdState->server_fd);
     assert(fd > -1);
+    commSetDefer(fd, NULL, NULL);
+    if (EBIT_TEST(fwdState->entry->flags, ENTRY_DEFER_READ))
+	storeResumeRead(fwdState->entry);
     comm_remove_close_handler(fd, fwdServerClosed, fwdState);
     fwdState->server_fd = -1;
 }
@@ -814,13 +1081,14 @@ int
 fwdReforwardableStatus(http_status s)
 {
     switch (s) {
+    case HTTP_BAD_GATEWAY:
+    case HTTP_GATEWAY_TIMEOUT:
+	return 1;
     case HTTP_FORBIDDEN:
     case HTTP_INTERNAL_SERVER_ERROR:
     case HTTP_NOT_IMPLEMENTED:
-    case HTTP_BAD_GATEWAY:
     case HTTP_SERVICE_UNAVAILABLE:
-    case HTTP_GATEWAY_TIMEOUT:
-	return 1;
+	return Config.retry.onerror;
     default:
 	return 0;
     }
