@@ -1,6 +1,6 @@
 
 /*
- * $Id: gopher.c,v 1.162.2.6 2002/09/01 12:38:04 hno Exp $
+ * $Id: gopher.c,v 1.177 2006/05/15 23:43:07 hno Exp $
  *
  * DEBUG: section 10    Gopher
  * AUTHOR: Harvest Derived
@@ -77,9 +77,9 @@ typedef struct gopher_ds {
 	HTML_CSO_PAGE
     } conversion;
     int HTML_header_added;
+    int HTML_pre;
     char type_id;
     char request[MAX_URL];
-    int data_in;
     int cso_recno;
     int len;
     char *buf;			/* pts to a 4k page */
@@ -257,7 +257,7 @@ gopherHTMLFooter(StoreEntry * e)
     storeAppendPrintf(e, "Generated %s by %s (%s)\n",
 	mkrfc1123(squid_curtime),
 	getMyHostname(),
-	full_appname_string);
+	visible_appname_string);
     storeAppendPrintf(e, "</ADDRESS></BODY></HTML>\n");
 }
 
@@ -265,10 +265,10 @@ static void
 gopherEndHTML(GopherStateData * gopherState)
 {
     StoreEntry *e = gopherState->entry;
-    if (!gopherState->data_in) {
+    if (!gopherState->HTML_header_added) {
 	gopherHTMLHeader(e, "Server Return Nothing", NULL);
 	storeAppendPrintf(e, "<P>The Gopher query resulted in a blank response</P>");
-    } else {
+    } else if (gopherState->HTML_pre) {
 	storeAppendPrintf(e, "</PRE>\n");
     }
     gopherHTMLFooter(e);
@@ -310,8 +310,7 @@ gopherToHTML(GopherStateData * gopherState, char *inbuf, int len)
 	gopherHTMLFooter(entry);
 	/* now let start sending stuff to client */
 	storeBufferFlush(entry);
-	gopherState->data_in = 1;
-
+	gopherState->HTML_header_added = 1;
 	return;
     }
     if (gopherState->conversion == HTML_CSO_PAGE) {
@@ -324,8 +323,7 @@ gopherToHTML(GopherStateData * gopherState, char *inbuf, int len)
 	gopherHTMLFooter(entry);
 	/* now let start sending stuff to client */
 	storeBufferFlush(entry);
-	gopherState->data_in = 1;
-
+	gopherState->HTML_header_added = 1;
 	return;
     }
     inbuf[len] = '\0';
@@ -337,12 +335,18 @@ gopherToHTML(GopherStateData * gopherState, char *inbuf, int len)
 	    gopherHTMLHeader(entry, "Gopher Menu", NULL);
 	strCat(outbuf, "<PRE>");
 	gopherState->HTML_header_added = 1;
+	gopherState->HTML_pre = 1;
     }
     while ((pos != NULL) && (pos < inbuf + len)) {
 
 	if (gopherState->len != 0) {
 	    /* there is something left from last tx. */
 	    xstrncpy(line, gopherState->buf, gopherState->len + 1);
+	    if (gopherState->len + len > TEMP_BUF_SIZE) {
+		debug(10, 1) ("GopherHTML: Buffer overflow. Lost some data on URL: %s\n",
+		    storeUrl(entry));
+		len = TEMP_BUF_SIZE - gopherState->len;
+	    }
 	    lpos = (char *) memccpy(line + gopherState->len, inbuf, '\n', len);
 	    if (lpos)
 		*lpos = '\0';
@@ -505,7 +509,6 @@ gopherToHTML(GopherStateData * gopherState, char *inbuf, int len)
 			}
 			safe_free(escaped_selector);
 			strCat(outbuf, tmpbuf);
-			gopherState->data_in = 1;
 		    } else {
 			memset(line, '\0', TEMP_BUF_SIZE);
 			continue;
@@ -543,7 +546,6 @@ gopherToHTML(GopherStateData * gopherState, char *inbuf, int len)
 			snprintf(tmpbuf, TEMP_BUF_SIZE, "%s\n", html_quote(result));
 		    }
 		    strCat(outbuf, tmpbuf);
-		    gopherState->data_in = 1;
 		    break;
 		} else {
 		    int code;
@@ -571,7 +573,6 @@ gopherToHTML(GopherStateData * gopherState, char *inbuf, int len)
 			    /* Print the message the server returns */
 			    snprintf(tmpbuf, TEMP_BUF_SIZE, "</PRE><HR noshade size=\"1px\"><H2>%s</H2>\n<PRE>", html_quote(result));
 			    strCat(outbuf, tmpbuf);
-			    gopherState->data_in = 1;
 			    break;
 			}
 
@@ -602,12 +603,8 @@ gopherTimeout(int fd, void *data)
     GopherStateData *gopherState = data;
     StoreEntry *entry = gopherState->entry;
     debug(10, 4) ("gopherTimeout: FD %d: '%s'\n", fd, storeUrl(entry));
-    if (entry->store_status == STORE_PENDING) {
-	if (entry->mem_obj->inmem_hi == 0) {
-	    fwdFail(gopherState->fwdState,
-		errorCon(ERR_READ_TIMEOUT, HTTP_GATEWAY_TIMEOUT));
-	}
-    }
+    fwdFail(gopherState->fwdState,
+	errorCon(ERR_READ_TIMEOUT, HTTP_GATEWAY_TIMEOUT));
     comm_close(fd);
 }
 
@@ -624,7 +621,7 @@ gopherReadReply(int fd, void *data)
     int bin;
     size_t read_sz;
 #if DELAY_POOLS
-    delay_id delay_id = delayMostBytesAllowed(entry->mem_obj);
+    delay_id delay_id;
 #endif
     if (EBIT_TEST(entry->flags, ENTRY_ABORTED)) {
 	comm_close(fd);
@@ -634,7 +631,7 @@ gopherReadReply(int fd, void *data)
     buf = memAllocate(MEM_4K_BUF);
     read_sz = 4096 - 1;		/* leave room for termination */
 #if DELAY_POOLS
-    read_sz = delayBytesWanted(delay_id, 1, read_sz);
+    delay_id = delayMostBytesAllowed(entry->mem_obj, &read_sz);
 #endif
     /* leave one space for \0 in gopherToHTML */
     statCounter.syscalls.sock.reads++;
@@ -659,22 +656,15 @@ gopherReadReply(int fd, void *data)
 	debug(50, 1) ("gopherReadReply: error reading: %s\n", xstrerror());
 	if (ignoreErrno(errno)) {
 	    commSetSelect(fd, COMM_SELECT_READ, gopherReadReply, data, 0);
-	} else if (entry->mem_obj->inmem_hi == 0) {
+	} else {
 	    ErrorState *err;
 	    err = errorCon(ERR_READ_ERROR, HTTP_INTERNAL_SERVER_ERROR);
 	    err->xerrno = errno;
-	    err->url = xstrdup(storeUrl(entry));
-	    errorAppendEntry(entry, err);
-	    comm_close(fd);
-	} else {
+	    fwdFail(gopherState->fwdState, err);
 	    comm_close(fd);
 	}
     } else if (len == 0 && entry->mem_obj->inmem_hi == 0) {
-	ErrorState *err;
-	err = errorCon(ERR_ZERO_SIZE_OBJECT, HTTP_SERVICE_UNAVAILABLE);
-	err->xerrno = errno;
-	err->url = xstrdup(gopherState->request);
-	errorAppendEntry(entry, err);
+	fwdFail(gopherState->fwdState, errorCon(ERR_ZERO_SIZE_OBJECT, HTTP_SERVICE_UNAVAILABLE));
 	comm_close(fd);
     } else if (len == 0) {
 	/* Connection closed; retrieval done. */
@@ -716,12 +706,10 @@ gopherSendComplete(int fd, char *buf, size_t size, int errflag, void *data)
     }
     if (errflag) {
 	ErrorState *err;
-	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE);
+	err = errorCon(ERR_WRITE_ERROR, HTTP_BAD_GATEWAY);
 	err->xerrno = errno;
-	err->host = xstrdup(gopherState->req->host);
-	err->port = gopherState->req->port;
 	err->url = xstrdup(storeUrl(entry));
-	errorAppendEntry(entry, err);
+	fwdFail(gopherState->fwdState, err);
 	comm_close(fd);
 	if (buf)
 	    memFree(buf, MEM_4K_BUF);	/* Allocated by gopherSendRequest. */
@@ -731,29 +719,28 @@ gopherSendComplete(int fd, char *buf, size_t size, int errflag, void *data)
      * OK. We successfully reach remote site.  Start MIME typing
      * stuff.  Do it anyway even though request is not HTML type.
      */
+    storeBuffer(entry);
     gopherMimeCreate(gopherState);
     switch (gopherState->type_id) {
     case GOPHER_DIRECTORY:
 	/* we got to convert it first */
-	storeBuffer(entry);
 	gopherState->conversion = HTML_DIR;
 	gopherState->HTML_header_added = 0;
 	break;
     case GOPHER_INDEX:
 	/* we got to convert it first */
-	storeBuffer(entry);
 	gopherState->conversion = HTML_INDEX_RESULT;
 	gopherState->HTML_header_added = 0;
 	break;
     case GOPHER_CSO:
 	/* we got to convert it first */
-	storeBuffer(entry);
 	gopherState->conversion = HTML_CSO_RESULT;
 	gopherState->cso_recno = 0;
 	gopherState->HTML_header_added = 0;
 	break;
     default:
 	gopherState->conversion = NORMAL;
+	storeBufferFlush(entry);
     }
     /* Schedule read reply. */
     commSetSelect(fd, COMM_SELECT_READ, gopherReadReply, gopherState, 0);
@@ -814,16 +801,6 @@ gopherStart(FwdState * fwdState)
     /* Parse url. */
     gopher_request_parse(fwdState->request,
 	&gopherState->type_id, gopherState->request);
-#if OLD_PARSE_ERROR_CODE
-    if (...) {
-	ErrorState *err;
-	err = errorCon(ERR_INVALID_URL, HTTP_BAD_REQUEST);
-	err->url = xstrdup(storeUrl(entry));
-	errorAppendEntry(entry, err);
-	gopherStateFree(-1, gopherState);
-	return;
-    }
-#endif
     comm_add_close_handler(fd, gopherStateFree, gopherState);
     if (((gopherState->type_id == GOPHER_INDEX) || (gopherState->type_id == GOPHER_CSO))
 	&& (strchr(gopherState->request, '?') == NULL)) {
