@@ -1,6 +1,6 @@
 
 /*
- * $Id: helper.c,v 1.34.2.12 2003/01/08 20:58:54 wessels Exp $
+ * $Id: helper.c,v 1.61 2006/07/09 15:44:32 serassio Exp $
  *
  * DEBUG: section 84    Helper process maintenance
  * AUTHOR: Harvest Derived?
@@ -53,9 +53,6 @@ static void helperStatefulKickQueue(statefulhelper * hlp);
 static void helperRequestFree(helper_request * r);
 static void helperStatefulRequestFree(helper_stateful_request * r);
 static void StatefulEnqueue(statefulhelper * hlp, helper_stateful_request * r);
-static helper_stateful_request *StatefulServerDequeue(helper_stateful_server * srv);
-static void StatefulServerEnqueue(helper_stateful_server * srv, helper_stateful_request * r);
-static void helperStatefulServerKickQueue(helper_stateful_server * srv);
 
 void
 helperOpenServers(helper * hlp)
@@ -103,15 +100,15 @@ helperOpenServers(helper * hlp)
 	    continue;
 	}
 	hlp->n_running++;
+	hlp->n_active++;
 	srv = cbdataAlloc(helper_server);
 	srv->pid = x;
-	srv->flags.alive = 1;
 	srv->index = k;
 	srv->rfd = rfd;
 	srv->wfd = wfd;
-	srv->buf = memAllocate(MEM_8K_BUF);
-	srv->buf_sz = 8192;
-	srv->offset = 0;
+	srv->rbuf = memAllocBuf(8192, &srv->rbuf_sz);
+	srv->roffset = 0;
+	srv->requests = xcalloc(hlp->concurrency ? hlp->concurrency : 1, sizeof(*srv->requests));
 	srv->parent = hlp;
 	cbdataLock(hlp);	/* lock because of the parent backlink */
 	dlinkAddTail(srv, &srv->link, &hlp->servers);
@@ -128,7 +125,9 @@ helperOpenServers(helper * hlp)
 	if (wfd != rfd)
 	    commSetNonBlocking(wfd);
 	comm_add_close_handler(rfd, helperServerFree, srv);
+	commSetSelect(srv->rfd, COMM_SELECT_READ, helperHandleRead, srv, 0);
     }
+    hlp->last_restart = squid_curtime;
     safe_free(shortname);
     safe_free(procname);
     helperKickQueue(hlp);
@@ -180,15 +179,11 @@ helperStatefulOpenServers(statefulhelper * hlp)
 	    continue;
 	}
 	hlp->n_running++;
+	hlp->n_active++;
 	srv = cbdataAlloc(helper_stateful_server);
 	srv->pid = x;
-	srv->flags.alive = 1;
-	srv->flags.reserved = S_HELPER_FREE;
-	srv->deferred_requests = 0;
-	srv->stats.deferbyfunc = 0;
-	srv->stats.deferbycb = 0;
+	srv->flags.reserved = 0;
 	srv->stats.submits = 0;
-	srv->stats.releases = 0;
 	srv->index = k;
 	srv->rfd = rfd;
 	srv->wfd = wfd;
@@ -213,7 +208,9 @@ helperStatefulOpenServers(statefulhelper * hlp)
 	if (wfd != rfd)
 	    commSetNonBlocking(wfd);
 	comm_add_close_handler(rfd, helperStatefulServerFree, srv);
+	commSetSelect(srv->rfd, COMM_SELECT_READ, helperStatefulHandleRead, srv, 0);
     }
+    hlp->last_restart = squid_curtime;
     safe_free(shortname);
     safe_free(procname);
     helperStatefulKickQueue(hlp);
@@ -241,14 +238,10 @@ helperSubmit(helper * hlp, const char *buf, HLPCB * callback, void *data)
     debug(84, 9) ("helperSubmit: %s\n", buf);
 }
 
-/* lastserver = "server last used as part of a deferred or reserved
- * request sequence"
- */
 void
-helperStatefulSubmit(statefulhelper * hlp, const char *buf, HLPSCB * callback, void *data, helper_stateful_server * lastserver)
+helperStatefulSubmit(statefulhelper * hlp, const char *buf, HLPSCB * callback, void *data, helper_stateful_server * srv)
 {
     helper_stateful_request *r = memAllocate(MEM_HELPER_STATEFUL_REQUEST);
-    helper_stateful_server *srv;
     if (hlp == NULL) {
 	debug(84, 3) ("helperStatefulSubmit: hlp == NULL\n");
 	callback(data, 0, NULL);
@@ -256,109 +249,50 @@ helperStatefulSubmit(statefulhelper * hlp, const char *buf, HLPSCB * callback, v
     }
     r->callback = callback;
     r->data = data;
-    if (buf != NULL) {
+    if (buf)
 	r->buf = xstrdup(buf);
-	r->placeholder = 0;
-    } else {
-	r->buf = NULL;
-	r->placeholder = 1;
-    }
     cbdataLock(r->data);
-    if ((buf != NULL) && lastserver) {
-	debug(84, 5) ("StatefulSubmit with lastserver %p\n", lastserver);
-	/* the queue doesn't count for this assert because queued requests
-	 * have already gone through here and been tested.
-	 * It's legal to have deferred_requests == 0 and queue entries 
-	 * and status of S_HELPEER_DEFERRED.
-	 * BUT:  It's not legal to submit a new request w/lastserver in
-	 * that state.
-	 */
-	assert(!(lastserver->deferred_requests == 0 &&
-		lastserver->flags.reserved == S_HELPER_DEFERRED));
-	if (lastserver->flags.reserved != S_HELPER_RESERVED) {
-	    lastserver->stats.submits++;
-	    lastserver->deferred_requests--;
-	}
-	if (!(lastserver->request)) {
-	    debug(84, 5) ("StatefulSubmit dispatching\n");
-	    helperStatefulDispatch(lastserver, r);
-	} else {
-	    debug(84, 5) ("StatefulSubmit queuing\n");
-	    StatefulServerEnqueue(lastserver, r);
-	}
+    if (!srv)
+	srv = helperStatefulGetServer(hlp);
+    if (srv) {
+	debug(84, 5) ("helperStatefulSubmit: server %p, buf '%s'.\n", srv, buf ? buf : "NULL");
+	assert(!srv->request);
+	assert(!srv->flags.busy);
+	helperStatefulDispatch(srv, r);
     } else {
-	if ((srv = StatefulGetFirstAvailable(hlp))) {
-	    helperStatefulDispatch(srv, r);
-	} else
-	    StatefulEnqueue(hlp, r);
+	debug(84, 9) ("helperStatefulSubmit: enqueued, buf '%s'.\n", buf ? buf : "NULL");
+	StatefulEnqueue(hlp, r);
     }
-    debug(84, 9) ("helperStatefulSubmit: placeholder: '%d', buf '%s'.\n", r->placeholder, buf);
 }
 
 helper_stateful_server *
-helperStatefulDefer(statefulhelper * hlp)
-/* find and add a deferred request to a server */
+helperStatefulGetServer(statefulhelper * hlp)
+/* find a server for this request */
 {
-    dlink_node *n;
-    helper_stateful_server *srv = NULL, *rv = NULL;
+    helper_stateful_server *srv = NULL;
     if (hlp == NULL) {
-	debug(84, 3) ("helperStatefulDefer: hlp == NULL\n");
+	debug(84, 3) ("helperStatefulGetServer: hlp == NULL\n");
 	return NULL;
     }
-    debug(84, 5) ("helperStatefulDefer: Running servers %d.\n", hlp->n_running);
+    debug(84, 5) ("helperStatefulGetServer: Running servers %d.\n", hlp->n_running);
     if (hlp->n_running == 0) {
-	debug(84, 1) ("helperStatefulDefer: No running servers!. \n");
+	debug(84, 1) ("helperStatefulGetServer: No running servers!. \n");
 	return NULL;
     }
-    rv = srv = StatefulGetFirstAvailable(hlp);
-    if (rv == NULL) {
-	/*
-	 * all currently busy; loop through servers and find server
-	 * with the shortest queue
-	 */
-	for (n = hlp->servers.head; n != NULL; n = n->next) {
-	    srv = n->data;
-	    if (srv->flags.reserved == S_HELPER_RESERVED)
-		continue;
-	    if (!srv->flags.alive)
-		continue;
-	    if ((hlp->IsAvailable != NULL) && (srv->data != NULL) &&
-		!(hlp->IsAvailable(srv->data)))
-		continue;
-	    if ((rv != NULL) && (rv->deferred_requests < srv->deferred_requests))
-		continue;
-	    rv = srv;
-	}
-    }
-    if (rv == NULL) {
-	debug(84, 1) ("helperStatefulDefer: None available.\n");
-	return NULL;
-    }
-    /* consistency check:
-     * when the deferred count is 0,
-     *   submits + releases == deferbyfunc + deferbycb
-     * Or in english, when there are no deferred requests, the amount
-     * we have submitted to the queue or cancelled must equal the amount
-     * we have said we wanted to be able to submit or cancel
-     */
-    if (rv->deferred_requests == 0)
-	assert(rv->stats.submits + rv->stats.releases ==
-	    rv->stats.deferbyfunc + rv->stats.deferbycb);
-
-    rv->flags.reserved = S_HELPER_DEFERRED;
-    rv->deferred_requests++;
-    rv->stats.deferbyfunc++;
-    return rv;
+    srv = StatefulGetFirstAvailable(hlp);
+    if (srv)
+	srv->flags.reserved = 1;
+    debug(84, 5) ("helperStatefulGetServer: Returning %p\n", srv);
+    return srv;
 }
 
+/* puts this helper forcibly back in the queue. */
 void
 helperStatefulReset(helper_stateful_server * srv)
-/* puts this helper back in the queue. the calling app is required to 
- * manage the state in the helper.
- */
 {
     statefulhelper *hlp = srv->parent;
     helper_stateful_request *r;
+    debug(84, 5) ("helperStatefulReset: %p\n", srv);
     r = srv->request;
     if (r != NULL) {
 	/* reset attempt DURING an outstaning request */
@@ -370,31 +304,26 @@ helperStatefulReset(helper_stateful_server * srv)
 	srv->request = NULL;
     }
     srv->flags.busy = 0;
-    if (srv->queue.head) {
-	srv->flags.reserved = S_HELPER_DEFERRED;
-	helperStatefulServerKickQueue(srv);
+    srv->flags.reserved = 0;
+    if ((srv->parent->Reset != NULL) && (srv->data))
+	srv->parent->Reset(srv->data);
+    if (srv->flags.shutdown) {
+	int wfd = srv->wfd;
+	srv->wfd = -1;
+	comm_close(wfd);
     } else {
-	srv->flags.reserved = S_HELPER_FREE;
-	if ((srv->parent->OnEmptyQueue != NULL) && (srv->data))
-	    srv->parent->OnEmptyQueue(srv->data);
 	helperStatefulKickQueue(hlp);
     }
 }
 
+/* puts this helper back in the queue. */
 void
 helperStatefulReleaseServer(helper_stateful_server * srv)
-/*decrease the number of 'waiting' clients that set the helper to be DEFERRED */
 {
-    srv->stats.releases++;
-    if (srv->flags.reserved == S_HELPER_DEFERRED) {
-	assert(srv->deferred_requests);
-	srv->deferred_requests--;
-    }
-    if (!(srv->deferred_requests) && (srv->flags.reserved == S_HELPER_DEFERRED) && !(srv->queue.head)) {
-	srv->flags.reserved = S_HELPER_FREE;
-	if ((srv->parent->OnEmptyQueue != NULL) && (srv->data))
-	    srv->parent->OnEmptyQueue(srv->data);
-    }
+    debug(84, 5) ("helperStatefulReleaseServer: %p\n", srv);
+    assert(!srv->request);
+    assert(srv->flags.reserved);
+    helperStatefulReset(srv);
 }
 
 void *
@@ -407,7 +336,57 @@ helperStatefulServerGetData(helper_stateful_server * srv)
 void
 helperStats(StoreEntry * sentry, helper * hlp)
 {
-    helper_server *srv;
+    dlink_node *link;
+    storeAppendPrintf(sentry, "program: %s\n",
+	hlp->cmdline->key);
+    storeAppendPrintf(sentry, "number running: %d of %d\n",
+	hlp->n_running, hlp->n_to_start);
+    storeAppendPrintf(sentry, "requests sent: %d\n",
+	hlp->stats.requests);
+    storeAppendPrintf(sentry, "replies received: %d\n",
+	hlp->stats.replies);
+    storeAppendPrintf(sentry, "queue length: %d\n",
+	hlp->stats.queue_size);
+    storeAppendPrintf(sentry, "avg service time: %.2f msec\n",
+	(double) hlp->stats.avg_svc_time / 1000.0);
+    storeAppendPrintf(sentry, "\n");
+    storeAppendPrintf(sentry, "%7s\t%7s\t%7s\t%11s\t%9s\t%s\t%7s\t%7s\t%7s\n",
+	"#",
+	"FD",
+	"PID",
+	"# Requests",
+	"# Pending",
+	"Flags",
+	"Time",
+	"Offset",
+	"Request");
+    for (link = hlp->servers.head; link; link = link->next) {
+	helper_server *srv = link->data;
+	double tt = srv->requests[0] ? 0.001 *
+	tvSubMsec(srv->requests[0]->dispatch_time, current_time) : 0.0;
+	storeAppendPrintf(sentry, "%7d\t%7d\t%7d\t%11d\t%9d\t%c%c%c\t%7.3f\t%7d\t%s\n",
+	    srv->index + 1,
+	    srv->rfd,
+	    srv->pid,
+	    srv->stats.uses,
+	    srv->stats.pending,
+	    srv->stats.pending ? 'B' : ' ',
+	    srv->flags.closing ? 'C' : ' ',
+	    srv->flags.shutdown ? 'S' : ' ',
+	    tt < 0.0 ? 0.0 : tt,
+	    srv->roffset,
+	    srv->requests[0] ? log_quote(srv->requests[0]->buf) : "(none)");
+    }
+    storeAppendPrintf(sentry, "\nFlags key:\n\n");
+    storeAppendPrintf(sentry, "   B = BUSY\n");
+    storeAppendPrintf(sentry, "   C = CLOSING\n");
+    storeAppendPrintf(sentry, "   S = SHUTDOWN\n");
+}
+
+void
+helperStatefulStats(StoreEntry * sentry, statefulhelper * hlp)
+{
+    helper_stateful_server *srv;
     dlink_node *link;
     double tt;
     storeAppendPrintf(sentry, "program: %s\n",
@@ -441,74 +420,19 @@ helperStats(StoreEntry * sentry, helper * hlp)
 	    srv->rfd,
 	    srv->pid,
 	    srv->stats.uses,
-	    srv->flags.alive ? 'A' : ' ',
 	    srv->flags.busy ? 'B' : ' ',
 	    srv->flags.closing ? 'C' : ' ',
+	    srv->flags.reserved ? 'R' : ' ',
 	    srv->flags.shutdown ? 'S' : ' ',
 	    tt < 0.0 ? 0.0 : tt,
-	    (int) srv->offset,
+	    srv->offset,
 	    srv->request ? log_quote(srv->request->buf) : "(none)");
     }
     storeAppendPrintf(sentry, "\nFlags key:\n\n");
-    storeAppendPrintf(sentry, "   A = ALIVE\n");
-    storeAppendPrintf(sentry, "   B = BUSY\n");
-    storeAppendPrintf(sentry, "   C = CLOSING\n");
-    storeAppendPrintf(sentry, "   S = SHUTDOWN\n");
-}
-
-void
-helperStatefulStats(StoreEntry * sentry, statefulhelper * hlp)
-{
-    helper_stateful_server *srv;
-    dlink_node *link;
-    double tt;
-    storeAppendPrintf(sentry, "number running: %d of %d\n",
-	hlp->n_running, hlp->n_to_start);
-    storeAppendPrintf(sentry, "requests sent: %d\n",
-	hlp->stats.requests);
-    storeAppendPrintf(sentry, "replies received: %d\n",
-	hlp->stats.replies);
-    storeAppendPrintf(sentry, "queue length: %d\n",
-	hlp->stats.queue_size);
-    storeAppendPrintf(sentry, "avg service time: %d msec\n",
-	hlp->stats.avg_svc_time);
-    storeAppendPrintf(sentry, "\n");
-    storeAppendPrintf(sentry, "%7s\t%7s\t%7s\t%11s\t%20s\t%s\t%7s\t%7s\t%7s\n",
-	"#",
-	"FD",
-	"PID",
-	"# Requests",
-	"# Deferred Requests",
-	"Flags",
-	"Time",
-	"Offset",
-	"Request");
-    for (link = hlp->servers.head; link; link = link->next) {
-	srv = link->data;
-	tt = 0.001 * tvSubMsec(srv->dispatch_time, current_time);
-	storeAppendPrintf(sentry, "%7d\t%7d\t%7d\t%11d\t%20d\t%c%c%c%c%c%c\t%7.3f\t%7d\t%s\n",
-	    srv->index + 1,
-	    srv->rfd,
-	    srv->pid,
-	    srv->stats.uses,
-	    (int) srv->deferred_requests,
-	    srv->flags.alive ? 'A' : ' ',
-	    srv->flags.busy ? 'B' : ' ',
-	    srv->flags.closing ? 'C' : ' ',
-	    srv->flags.reserved != S_HELPER_FREE ? 'R' : ' ',
-	    srv->flags.shutdown ? 'S' : ' ',
-	    srv->request ? (srv->request->placeholder ? 'P' : ' ') : ' ',
-	    tt < 0.0 ? 0.0 : tt,
-	    (int) srv->offset,
-	    srv->request ? log_quote(srv->request->buf) : "(none)");
-    }
-    storeAppendPrintf(sentry, "\nFlags key:\n\n");
-    storeAppendPrintf(sentry, "   A = ALIVE\n");
     storeAppendPrintf(sentry, "   B = BUSY\n");
     storeAppendPrintf(sentry, "   C = CLOSING\n");
     storeAppendPrintf(sentry, "   R = RESERVED or DEFERRED\n");
     storeAppendPrintf(sentry, "   S = SHUTDOWN\n");
-    storeAppendPrintf(sentry, "   P = PLACEHOLDER\n");
 }
 
 void
@@ -520,13 +444,15 @@ helperShutdown(helper * hlp)
 	helper_server *srv;
 	srv = link->data;
 	link = link->next;
-	if (!srv->flags.alive) {
-	    debug(84, 3) ("helperShutdown: %s #%d is NOT ALIVE.\n",
+	if (srv->flags.shutdown) {
+	    debug(84, 3) ("helperShutdown: %s #%d is already shut down\n",
 		hlp->id_name, srv->index + 1);
 	    continue;
 	}
+	hlp->n_active--;
+	assert(hlp->n_active >= 0);
 	srv->flags.shutdown = 1;	/* request it to shut itself down */
-	if (srv->flags.busy) {
+	if (srv->flags.writing) {
 	    debug(84, 3) ("helperShutdown: %s #%d is BUSY.\n",
 		hlp->id_name, srv->index + 1);
 	    continue;
@@ -536,6 +462,11 @@ helperShutdown(helper * hlp)
 		hlp->id_name, srv->index + 1);
 	    continue;
 	}
+	if (srv->stats.pending) {
+	    debug(84, 3) ("helperShutdown: %s #%d is BUSY.\n",
+		hlp->id_name, srv->index + 1);
+	    continue;
+	}
 	srv->flags.closing = 1;
 	wfd = srv->wfd;
 	srv->wfd = -1;
@@ -552,11 +483,13 @@ helperStatefulShutdown(statefulhelper * hlp)
     while (link) {
 	srv = link->data;
 	link = link->next;
-	if (!srv->flags.alive) {
-	    debug(84, 3) ("helperStatefulShutdown: %s #%d is NOT ALIVE.\n",
+	if (srv->flags.shutdown) {
+	    debug(84, 3) ("helperStatefulShutdown: %s #%d is already shut down.\n",
 		hlp->id_name, srv->index + 1);
 	    continue;
 	}
+	hlp->n_active--;
+	assert(hlp->n_active >= 0);
 	srv->flags.shutdown = 1;	/* request it to shut itself down */
 	if (srv->flags.busy) {
 	    debug(84, 3) ("helperStatefulShutdown: %s #%d is BUSY.\n",
@@ -568,16 +501,11 @@ helperStatefulShutdown(statefulhelper * hlp)
 		hlp->id_name, srv->index + 1);
 	    continue;
 	}
-	if (srv->flags.reserved != S_HELPER_FREE) {
+	if (srv->flags.reserved) {
 	    debug(84, 3) ("helperStatefulShutdown: %s #%d is RESERVED.\n",
 		hlp->id_name, srv->index + 1);
 	    continue;
 	}
-	if (srv->deferred_requests) {
-	    debug(84, 3) ("helperStatefulShutdown: %s #%d has DEFERRED requests.\n",
-		hlp->id_name, srv->index + 1);
-	    continue;
-	}
 	srv->flags.closing = 1;
 	wfd = srv->wfd;
 	srv->wfd = -1;
@@ -640,27 +568,42 @@ helperServerFree(int fd, void *data)
     helper_server *srv = data;
     helper *hlp = srv->parent;
     helper_request *r;
+    int i, concurrency = hlp->concurrency;
+    if (!concurrency)
+	concurrency = 1;
     assert(srv->rfd == fd);
-    if (srv->buf) {
-	memFree(srv->buf, MEM_8K_BUF);
-	srv->buf = NULL;
+    if (srv->rbuf) {
+	memFreeBuf(srv->rbuf_sz, srv->rbuf);
+	srv->rbuf = NULL;
     }
-    if ((r = srv->request)) {
-	if (cbdataValid(r->data))
-	    r->callback(r->data, srv->buf);
-	helperRequestFree(r);
-	srv->request = NULL;
+    if (!memBufIsNull(&srv->wqueue))
+	memBufClean(&srv->wqueue);
+    for (i = 0; i < concurrency; i++) {
+	if ((r = srv->requests[i])) {
+	    if (cbdataValid(r->data))
+		r->callback(r->data, NULL);
+	    helperRequestFree(r);
+	    srv->requests[i] = NULL;
+	}
     }
+    safe_free(srv->requests);
     if (srv->wfd != srv->rfd && srv->wfd != -1)
 	comm_close(srv->wfd);
     dlinkDelete(&srv->link, &hlp->servers);
     hlp->n_running--;
     assert(hlp->n_running >= 0);
     if (!srv->flags.shutdown) {
+	hlp->n_active--;
+	assert(hlp->n_active >= 0);
 	debug(84, 0) ("WARNING: %s #%d (FD %d) exited\n",
 	    hlp->id_name, srv->index + 1, fd);
-	if (hlp->n_running < hlp->n_to_start / 2)
-	    fatalf("Too few %s processes are running", hlp->id_name);
+	if (hlp->n_active <= hlp->n_to_start / 2) {
+	    debug(84, 0) ("Too few %s processes are running\n", hlp->id_name);
+	    if (hlp->last_restart > squid_curtime - 30)
+		fatalf("The %s helpers are crashing too rapidly, need help!\n", hlp->id_name);
+	    debug(84, 0) ("Starting new helpers\n");
+	    helperOpenServers(hlp);
+	}
     }
     cbdataUnlock(srv->parent);
     cbdataFree(srv);
@@ -690,10 +633,17 @@ helperStatefulServerFree(int fd, void *data)
     hlp->n_running--;
     assert(hlp->n_running >= 0);
     if (!srv->flags.shutdown) {
+	hlp->n_active--;
+	assert(hlp->n_active >= 0);
 	debug(84, 0) ("WARNING: %s #%d (FD %d) exited\n",
 	    hlp->id_name, srv->index + 1, fd);
-	if (hlp->n_running < hlp->n_to_start / 2)
-	    fatalf("Too few %s processes are running", hlp->id_name);
+	if (hlp->n_active <= hlp->n_to_start / 2) {
+	    debug(84, 0) ("Too few %s processes are running\n", hlp->id_name);
+	    if (hlp->last_restart > squid_curtime - 30)
+		fatalf("The %s helpers are crashing too rapidly, need help!\n", hlp->id_name);
+	    debug(84, 0) ("Starting new helpers\n");
+	    helperStatefulOpenServers(hlp);
+	}
     }
     if (srv->data != NULL)
 	memPoolFree(hlp->datapool, srv->data);
@@ -708,53 +658,80 @@ helperHandleRead(int fd, void *data)
     int len;
     char *t = NULL;
     helper_server *srv = data;
-    helper_request *r;
     helper *hlp = srv->parent;
     assert(fd == srv->rfd);
     assert(cbdataValid(data));
     statCounter.syscalls.sock.reads++;
-    len = FD_READ_METHOD(fd, srv->buf + srv->offset, srv->buf_sz - srv->offset);
+    /* XXX srv->rbuf should be reallocated if needed.. and start out quite small (not fixed 8KB as now..) */
+    assert(srv->roffset < srv->rbuf_sz);
+    len = FD_READ_METHOD(fd, srv->rbuf + srv->roffset, srv->rbuf_sz - srv->roffset - 1);
     fd_bytes(fd, len, FD_READ);
     debug(84, 5) ("helperHandleRead: %d bytes from %s #%d.\n",
 	len, hlp->id_name, srv->index + 1);
-    if (len <= 0) {
-	if (len < 0)
-	    debug(84, 1) ("helperHandleRead: FD %d read: %s\n", fd, xstrerror());
+    if (len == 0) {
 	comm_close(fd);
 	return;
     }
-    srv->offset += len;
-    srv->buf[srv->offset] = '\0';
-    r = srv->request;
-    if (r == NULL) {
+    commSetSelect(fd, COMM_SELECT_READ, helperHandleRead, srv, 0);
+    if (len < 0) {
+	if (!ignoreErrno(errno)) {
+	    debug(84, 1) ("helperHandleRead: FD %d read: %s\n", fd, xstrerror());
+	    comm_close(fd);
+	}
+	return;
+    }
+    srv->roffset += len;
+    srv->rbuf[srv->roffset] = '\0';
+    debug(84, 9) ("helperHandleRead: '%s'\n", srv->rbuf);
+    if (!srv->stats.pending) {
 	/* someone spoke without being spoken to */
-	debug(84, 1) ("helperHandleRead: unexpected read from %s #%d, %d bytes\n",
-	    hlp->id_name, srv->index + 1, len);
-	srv->offset = 0;
-    } else if ((t = strchr(srv->buf, '\n'))) {
+	debug(84, 1) ("helperHandleRead: unexpected read from %s #%d, %d bytes '%s'\n",
+	    hlp->id_name, srv->index + 1, len, srv->rbuf);
+	srv->roffset = 0;
+	srv->rbuf[0] = '\0';
+    }
+    while ((t = strchr(srv->rbuf, '\n'))) {
+	helper_request *r;
+	char *msg = srv->rbuf;
+	int i = 0;
 	/* end of reply found */
-	debug(84, 3) ("helperHandleRead: end of reply found\n");
-	*t = '\0';
-	if (cbdataValid(r->data))
-	    r->callback(r->data, srv->buf);
-	srv->flags.busy = 0;
-	srv->offset = 0;
-	helperRequestFree(r);
-	srv->request = NULL;
-	hlp->stats.replies++;
-	srv->answer_time = current_time;
-	hlp->stats.avg_svc_time =
-	    intAverage(hlp->stats.avg_svc_time,
-	    tvSubUsec(srv->dispatch_time, current_time),
-	    hlp->stats.replies, REDIRECT_AV_FACTOR);
-	if (srv->flags.shutdown) {
+	debug(84, 3) ("helperHandleRead: end of reply found: %s\n", srv->rbuf);
+	if (t > srv->rbuf && t[-1] == '\r')
+	    t[-1] = '\0';
+	*t++ = '\0';
+	if (hlp->concurrency) {
+	    i = strtol(msg, &msg, 10);
+	    while (*msg && isspace((int) *msg))
+		msg++;
+	}
+	r = srv->requests[i];
+	if (r) {
+	    srv->requests[i] = NULL;
+	    if (cbdataValid(r->data))
+		r->callback(r->data, msg);
+	    srv->stats.pending--;
+	    hlp->stats.replies++;
+	    hlp->stats.avg_svc_time =
+		intAverage(hlp->stats.avg_svc_time,
+		tvSubUsec(r->dispatch_time, current_time),
+		hlp->stats.replies, REDIRECT_AV_FACTOR);
+	    helperRequestFree(r);
+	} else {
+	    debug(84, 1) ("helperHandleRead: unexpected reply on channel %d from %s #%d '%s'\n",
+		i, hlp->id_name, srv->index + 1, srv->rbuf);
+	}
+	srv->roffset -= (t - srv->rbuf);
+	memmove(srv->rbuf, t, srv->roffset + 1);
+    }
+    if (srv->flags.shutdown && !srv->stats.pending) {
+	if (!srv->flags.closing) {
 	    int wfd = srv->wfd;
+	    srv->flags.closing = 1;
 	    srv->wfd = -1;
 	    comm_close(wfd);
-	} else
-	    helperKickQueue(hlp);
+	}
     } else {
-	commSetSelect(srv->rfd, COMM_SELECT_READ, helperHandleRead, srv, 0);
+	helperKickQueue(hlp);
     }
 }
 
@@ -779,6 +756,7 @@ helperStatefulHandleRead(int fd, void *data)
 	comm_close(fd);
 	return;
     }
+    commSetSelect(srv->rfd, COMM_SELECT_READ, helperStatefulHandleRead, srv, 0);
     srv->offset += len;
     srv->buf[srv->offset] = '\0';
     r = srv->request;
@@ -790,71 +768,24 @@ helperStatefulHandleRead(int fd, void *data)
     } else if ((t = strchr(srv->buf, '\n'))) {
 	/* end of reply found */
 	debug(84, 3) ("helperStatefulHandleRead: end of reply found\n");
+	if (t > srv->buf && t[-1] == '\r')
+	    t[-1] = '\0';
 	*t = '\0';
+	srv->flags.busy = 0;
+	srv->offset = 0;
+	srv->request = NULL;
+	hlp->stats.replies++;
+	srv->answer_time = current_time;
+	hlp->stats.avg_svc_time =
+	    intAverage(hlp->stats.avg_svc_time,
+	    tvSubUsec(srv->dispatch_time, current_time),
+	    hlp->stats.replies, REDIRECT_AV_FACTOR);
 	if (cbdataValid(r->data)) {
-	    switch ((r->callback(r->data, srv, srv->buf))) {	/*if non-zero reserve helper */
-	    case S_HELPER_UNKNOWN:
-		fatal("helperStatefulHandleRead: either a non-state aware callback was give to the stateful helper routines, or an uninitialised callback response was recieved.\n");
-		break;
-	    case S_HELPER_RELEASE:	/* helper finished with */
-		if (!srv->deferred_requests && !srv->queue.head) {
-		    srv->flags.reserved = S_HELPER_FREE;
-		    if ((srv->parent->OnEmptyQueue != NULL) && (srv->data))
-			srv->parent->OnEmptyQueue(srv->data);
-		    debug(84, 5) ("StatefulHandleRead: releasing %s #%d\n", hlp->id_name, srv->index + 1);
-		} else {
-		    srv->flags.reserved = S_HELPER_DEFERRED;
-		    debug(84, 5) ("StatefulHandleRead: outstanding deferred requests on %s #%d. reserving for deferred requests.\n", hlp->id_name, srv->index + 1);
-		}
-		break;
-	    case S_HELPER_RESERVE:	/* 'pin' this helper for the caller */
-		if (!srv->queue.head) {
-		    assert(srv->deferred_requests == 0);
-		    srv->flags.reserved = S_HELPER_RESERVED;
-		    debug(84, 5) ("StatefulHandleRead: reserving %s #%d\n", hlp->id_name, srv->index + 1);
-		} else {
-		    fatal("StatefulHandleRead: Callback routine attempted to reserve a stateful helper with deferred requests. This can lead to deadlock.\n");
-		}
-		break;
-	    case S_HELPER_DEFER:
-		/* the helper is still needed, but can
-		 * be used for other requests in the meantime.
-		 */
-		srv->flags.reserved = S_HELPER_DEFERRED;
-		srv->deferred_requests++;
-		srv->stats.deferbycb++;
-		debug(84, 5) ("StatefulHandleRead: reserving %s #%d for deferred requests.\n", hlp->id_name, srv->index + 1);
-		break;
-	    default:
-		fatal("helperStatefulHandleRead: unknown stateful helper callback result.\n");
-	    }
-
+	    r->callback(r->data, srv, srv->buf);
 	} else {
 	    debug(84, 1) ("StatefulHandleRead: no callback data registered\n");
 	}
-	srv->flags.busy = 0;
-	srv->offset = 0;
 	helperStatefulRequestFree(r);
-	srv->request = NULL;
-	hlp->stats.replies++;
-	hlp->stats.avg_svc_time =
-	    intAverage(hlp->stats.avg_svc_time,
-	    tvSubMsec(srv->dispatch_time, current_time),
-	    hlp->stats.replies, REDIRECT_AV_FACTOR);
-	if (srv->flags.shutdown
-	    && srv->flags.reserved == S_HELPER_FREE
-	    && !srv->deferred_requests) {
-	    int wfd = srv->wfd;
-	    srv->wfd = -1;
-	    comm_close(wfd);
-	} else {
-	    if (srv->queue.head)
-		helperStatefulServerKickQueue(srv);
-	    else
-		helperStatefulKickQueue(hlp);
-	}
-    } else {
-	commSetSelect(srv->rfd, COMM_SELECT_READ, helperStatefulHandleRead, srv, 0);
     }
 }
 
@@ -866,16 +797,20 @@ Enqueue(helper * hlp, helper_request * r)
     hlp->stats.queue_size++;
     if (hlp->stats.queue_size < hlp->n_running)
 	return;
-    if (squid_curtime - hlp->last_queue_warn < 600)
+    if (hlp->stats.queue_size > hlp->stats.max_queue_size)
+	hlp->stats.max_queue_size = hlp->stats.queue_size;
+    if (squid_curtime - hlp->last_queue_warn < 30)
 	return;
     if (shutting_down || reconfiguring)
 	return;
-    hlp->last_queue_warn = squid_curtime;
-    debug(84, 0) ("WARNING: All %s processes are busy.\n", hlp->id_name);
-    debug(84, 0) ("WARNING: %d pending requests queued\n", hlp->stats.queue_size);
+    debug(84, 1) ("WARNING: All %s processes are busy.\n", hlp->id_name);
+    debug(84, 1) ("WARNING: up to %d pending requests queued\n", hlp->stats.max_queue_size);
     if (hlp->stats.queue_size > hlp->n_running * 2)
-	fatalf("Too many queued %s requests", hlp->id_name);
-    debug(84, 1) ("Consider increasing the number of %s processes in your config file.\n", hlp->id_name);
+	fatalf("Too many queued %s requests (%d on %d)", hlp->id_name, hlp->stats.queue_size, hlp->n_running);
+    if (squid_curtime - hlp->last_queue_warn < 300)
+	debug(84, 1) ("Consider increasing the number of %s processes to at least %d in your config file.\n", hlp->id_name, hlp->n_running + hlp->stats.max_queue_size);
+    hlp->last_queue_warn = squid_curtime;
+    hlp->stats.max_queue_size = hlp->stats.queue_size;
 }
 
 static void
@@ -886,43 +821,22 @@ StatefulEnqueue(statefulhelper * hlp, helper_stateful_request * r)
     hlp->stats.queue_size++;
     if (hlp->stats.queue_size < hlp->n_running)
 	return;
-    if (hlp->stats.queue_size > hlp->n_running * 2)
-	fatalf("Too many queued %s requests", hlp->id_name);
-    if (squid_curtime - hlp->last_queue_warn < 600)
+    if (hlp->stats.queue_size > hlp->stats.max_queue_size)
+	hlp->stats.max_queue_size = hlp->stats.queue_size;
+    if (hlp->stats.queue_size > hlp->n_running * 5)
+	fatalf("Too many queued %s requests (%d on %d)", hlp->id_name, hlp->stats.queue_size, hlp->n_running);
+    if (squid_curtime - hlp->last_queue_warn < 30)
 	return;
     if (shutting_down || reconfiguring)
 	return;
+    debug(84, 1) ("WARNING: All %s processes are busy.\n", hlp->id_name);
+    debug(84, 1) ("WARNING: up to %d pending requests queued\n", hlp->stats.max_queue_size);
+    if (squid_curtime - hlp->last_queue_warn < 300)
+	debug(84, 1) ("Consider increasing the number of %s processes to at least %d in your config file.\n", hlp->id_name, hlp->n_running + hlp->stats.max_queue_size);
     hlp->last_queue_warn = squid_curtime;
-    debug(84, 0) ("WARNING: All %s processes are busy.\n", hlp->id_name);
-    debug(84, 0) ("WARNING: %d pending requests queued\n", hlp->stats.queue_size);
-    debug(84, 1) ("Consider increasing the number of %s processes in your config file.\n", hlp->id_name);
+    hlp->stats.max_queue_size = hlp->stats.queue_size;
 }
 
-static void
-StatefulServerEnqueue(helper_stateful_server * srv, helper_stateful_request * r)
-{
-    dlink_node *link = memAllocate(MEM_DLINK_NODE);
-    dlinkAddTail(r, link, &srv->queue);
-/* TODO: warning if the queue on this server is more than X
- * We don't check the queue size at the moment, because
- * requests hitting here are deferrable 
- */
-/*    hlp->stats.queue_size++;
- * if (hlp->stats.queue_size < hlp->n_running)
- * return;
- * if (squid_curtime - hlp->last_queue_warn < 600)
- * return;
- * if (shutting_down || reconfiguring)
- * return;
- * hlp->last_queue_warn = squid_curtime;
- * debug(84, 0) ("WARNING: All %s processes are busy.\n", hlp->id_name);
- * debug(84, 0) ("WARNING: %d pending requests queued\n", hlp->stats.queue_size);
- * if (hlp->stats.queue_size > hlp->n_running * 2)
- * fatalf("Too many queued %s requests", hlp->id_name);
- * debug(84, 1) ("Consider increasing the number of %s processes in your config file.\n", hlp->id_name);  */
-}
-
-
 static helper_request *
 Dequeue(helper * hlp)
 {
@@ -937,19 +851,6 @@ Dequeue(helper * hlp)
     return r;
 }
 
-static helper_stateful_request *
-StatefulServerDequeue(helper_stateful_server * srv)
-{
-    dlink_node *link;
-    helper_stateful_request *r = NULL;
-    if ((link = srv->queue.head)) {
-	r = link->data;
-	dlinkDelete(link, &srv->queue);
-	memFree(link, MEM_DLINK_NODE);
-    }
-    return r;
-}
-
 static helper_stateful_request *
 StatefulDequeue(statefulhelper * hlp)
 {
@@ -968,18 +869,34 @@ static helper_server *
 GetFirstAvailable(helper * hlp)
 {
     dlink_node *n;
-    helper_server *srv = NULL;
+    helper_server *srv;
+    helper_server *selected = NULL;
     if (hlp->n_running == 0)
 	return NULL;
+    /* Find "least" loaded helper (approx) */
     for (n = hlp->servers.head; n != NULL; n = n->next) {
 	srv = n->data;
-	if (srv->flags.busy)
+	if (selected && selected->stats.pending <= srv->stats.pending)
 	    continue;
-	if (!srv->flags.alive)
+	if (srv->flags.shutdown)
 	    continue;
-	return srv;
+	if (srv->flags.closing)
+	    continue;
+	if (selected) {
+	    selected = srv;
+	    break;
+	}
+	selected = srv;
+	if (!selected->stats.pending)
+	    break;
     }
-    return NULL;
+    /* Check for overload */
+    if (!selected)
+	return NULL;
+    if (selected->stats.pending >= (hlp->concurrency ? hlp->concurrency : 1))
+	return NULL;
+
+    return selected;
 }
 
 static helper_stateful_server *
@@ -994,9 +911,9 @@ StatefulGetFirstAvailable(statefulhelper * hlp)
 	srv = n->data;
 	if (srv->flags.busy)
 	    continue;
-	if (srv->flags.reserved == S_HELPER_RESERVED)
+	if (srv->flags.reserved)
 	    continue;
-	if (!srv->flags.alive)
+	if (srv->flags.shutdown)
 	    continue;
 	if ((hlp->IsAvailable != NULL) && (srv->data != NULL) && !(hlp->IsAvailable(srv->data)))
 	    continue;
@@ -1007,29 +924,81 @@ StatefulGetFirstAvailable(statefulhelper * hlp)
 }
 
 
+static void
+helperDispatch_done(int fd, char *buf, size_t size, int status, void *data)
+{
+    helper_server *srv = data;
+    if (status != COMM_OK) {
+	/* Helper server has crashed.. */
+	debug(84, 0) ("ERROR: Helper on fd %d has crashed!\n", fd);
+    } else if (!memBufIsNull(&srv->wqueue)) {
+	MemBuf mb = srv->wqueue;
+	srv->wqueue = MemBufNull;
+	comm_write_mbuf(srv->wfd,
+	    mb,
+	    helperDispatch_done,	/* Handler */
+	    srv);
+    } else {
+	helper *hlp = srv->parent;
+	srv->flags.writing = 0;	/* done */
+	if (srv->flags.shutdown) {
+	    int wfd;
+	    debug(84, 3) ("helperDispatch: %s #%d is shutting down.\n",
+		hlp->id_name, srv->index + 1);
+	    if (srv->flags.closing) {
+		debug(84, 3) ("helperDispatch: %s #%d is CLOSING.\n",
+		    hlp->id_name, srv->index + 1);
+		return;
+	    }
+	    if (srv->stats.pending) {
+		debug(84, 3) ("helperDispatch: %s #%d is BUSY.\n",
+		    hlp->id_name, srv->index + 1);
+		return;
+	    }
+	    srv->flags.closing = 1;
+	    wfd = srv->wfd;
+	    srv->wfd = -1;
+	    comm_close(wfd);
+	}
+    }
+}
+
 static void
 helperDispatch(helper_server * srv, helper_request * r)
 {
     helper *hlp = srv->parent;
+    helper_request **ptr = NULL;
+    int slot;
     if (!cbdataValid(r->data)) {
 	debug(84, 1) ("helperDispatch: invalid callback data\n");
 	helperRequestFree(r);
 	return;
     }
-    assert(!srv->flags.busy);
-    srv->flags.busy = 1;
-    srv->request = r;
-    srv->dispatch_time = current_time;
-    comm_write(srv->wfd,
-	r->buf,
-	strlen(r->buf),
-	NULL,			/* Handler */
-	NULL,			/* Handler-data */
-	NULL);			/* free */
-    commSetSelect(srv->rfd,
-	COMM_SELECT_READ,
-	helperHandleRead,
-	srv, 0);
+    for (slot = 0; slot < (hlp->concurrency ? hlp->concurrency : 1); slot++) {
+	if (!srv->requests[slot]) {
+	    ptr = &srv->requests[slot];
+	    break;
+	}
+    }
+    assert(ptr);
+    *ptr = r;
+    srv->stats.pending += 1;
+    r->dispatch_time = current_time;
+    if (memBufIsNull(&srv->wqueue))
+	memBufDefInit(&srv->wqueue);
+    if (hlp->concurrency)
+	memBufPrintf(&srv->wqueue, "%d %s", slot, r->buf);
+    else
+	memBufAppend(&srv->wqueue, r->buf, strlen(r->buf));
+    if (!srv->flags.writing) {
+	MemBuf mb = srv->wqueue;
+	srv->wqueue = MemBufNull;
+	srv->flags.writing = 1;
+	comm_write_mbuf(srv->wfd,
+	    mb,
+	    helperDispatch_done,	/* Handler */
+	    srv);
+    }
     debug(84, 5) ("helperDispatch: Request sent to %s #%d, %d bytes\n",
 	hlp->id_name, srv->index + 1, (int) strlen(r->buf));
     srv->stats.uses++;
@@ -1045,32 +1014,16 @@ helperStatefulDispatch(helper_stateful_server * srv, helper_stateful_request * r
 	helperStatefulRequestFree(r);
 	return;
     }
-    debug(84, 9) ("helperStatefulDispatch busying helper %s #%d\n", hlp->id_name, srv->index + 1);
-    if (r->placeholder == 1) {
-	/* a callback is needed before this request can _use_ a helper. */
-	/* we don't care about releasing/deferring this helper. The request NEVER
-	 * gets to the helper. So we throw away the return code */
-	r->callback(r->data, srv, NULL);
-	/* throw away the placeholder */
-	helperStatefulRequestFree(r);
-	/* and push the queue. Note that the callback may have submitted a new 
-	 * request to the helper which is why we test for the request*/
-	if (srv->request == NULL) {
-	    if (srv->flags.shutdown
-		&& srv->flags.reserved == S_HELPER_FREE
-		&& !srv->deferred_requests) {
-		int wfd = srv->wfd;
-		srv->wfd = -1;
-		comm_close(wfd);
-	    } else {
-		if (srv->queue.head)
-		    helperStatefulServerKickQueue(srv);
-		else
-		    helperStatefulKickQueue(hlp);
-	    }
+    if (!r->buf) {
+	if (cbdataValid(r->data)) {
+	    r->callback(r->data, srv, NULL);
+	} else {
+	    debug(84, 1) ("helperStatefulDispatch: no callback data registered\n");
 	}
+	helperStatefulRequestFree(r);
 	return;
     }
+    debug(84, 9) ("helperStatefulDispatch busying helper %s #%d\n", hlp->id_name, srv->index + 1);
     srv->flags.busy = 1;
     srv->request = r;
     srv->dispatch_time = current_time;
@@ -1080,10 +1033,6 @@ helperStatefulDispatch(helper_stateful_server * srv, helper_stateful_request * r
 	NULL,			/* Handler */
 	NULL,			/* Handler-data */
 	NULL);			/* free */
-    commSetSelect(srv->rfd,
-	COMM_SELECT_READ,
-	helperStatefulHandleRead,
-	srv, 0);
     debug(84, 5) ("helperStatefulDispatch: Request sent to %s #%d, %d bytes\n",
 	hlp->id_name, srv->index + 1, (int) strlen(r->buf));
     srv->stats.uses++;
@@ -1105,16 +1054,10 @@ helperStatefulKickQueue(statefulhelper * hlp)
 {
     helper_stateful_request *r;
     helper_stateful_server *srv;
-    while ((srv = StatefulGetFirstAvailable(hlp)) && (r = StatefulDequeue(hlp)))
-	helperStatefulDispatch(srv, r);
-}
-
-static void
-helperStatefulServerKickQueue(helper_stateful_server * srv)
-{
-    helper_stateful_request *r;
-    if ((r = StatefulServerDequeue(srv)))
+    while ((srv = StatefulGetFirstAvailable(hlp)) && (r = StatefulDequeue(hlp))) {
+	srv->flags.reserved = 1;
 	helperStatefulDispatch(srv, r);
+    }
 }
 
 static void
